@@ -678,27 +678,58 @@ static int _psl_idna_toASCII(_psl_idna_t *idna _UNUSED, const char *utf8, char *
 #if defined(WITH_LIBICU)
 	/* IDNA2008 UTS#46 punycode conversion */
 	if (idna) {
-		char lookupname[128] = "";
+		char lookupname_buf[128] = "", *lookupname = lookupname_buf;
 		UErrorCode status = 0;
 		UIDNAInfo info = UIDNA_INFO_INITIALIZER;
-		UChar utf16_dst[128], utf16_src[128];
-		int32_t utf16_src_length;
+		UChar utf16_dst[128], utf16_src_buf[128];
+		UChar *utf16_src = utf16_src_buf;
+		int32_t utf16_src_length, bytes_written;
+		int32_t utf16_dst_length;
 
-		u_strFromUTF8(utf16_src, countof(utf16_src), &utf16_src_length, utf8, -1, &status);
-		if (U_SUCCESS(status)) {
-			int32_t dst_length = uidna_nameToASCII((UIDNA *)idna, utf16_src, utf16_src_length, utf16_dst, countof(utf16_dst), &info, &status);
-			if (U_SUCCESS(status)) {
-				u_strToUTF8(lookupname, sizeof(lookupname), NULL, utf16_dst, dst_length, &status);
-				if (U_SUCCESS(status)) {
-					if (ascii)
-						if ((*ascii = strdup(lookupname)))
-							ret = 0;
-				} /* else
-					fprintf(stderr, "Failed to convert UTF-16 to UTF-8 (status %d)\n", status); */
-			} /* else
-				fprintf(stderr, "Failed to convert to ASCII (status %d)\n", status); */
-		} /* else
-			fprintf(stderr, "Failed to convert UTF-8 to UTF-16 (status %d)\n", status); */
+		u_strFromUTF8(utf16_src, countof(utf16_src_buf), &utf16_src_length, utf8, -1, &status);
+		if (!U_SUCCESS(status)) goto cleanup; /* UTF-8 to UTF-16 conversion failed */
+
+		if (utf16_src_length >= (int) countof(utf16_src_buf)) {
+			utf16_src = malloc((utf16_src_length + 1) * sizeof(UChar));
+			if (!utf16_src) goto cleanup;
+
+			u_strFromUTF8(utf16_src, utf16_src_length, NULL, utf8, -1, &status);
+			if (!U_SUCCESS(status)) goto cleanup; /* UTF-8 to UTF-16 conversion failed */
+
+			utf16_src[utf16_src_length] = 0; /* u_strFromUTF8() doesn't 0-terminate if dest is filled up */
+		}
+
+		utf16_dst_length = uidna_nameToASCII((UIDNA *)idna, utf16_src, utf16_src_length, utf16_dst, countof(utf16_dst), &info, &status);
+		if (!U_SUCCESS(status)) goto cleanup; /* to ASCII conversion failed */
+
+		u_strToUTF8(lookupname, sizeof(lookupname_buf), &bytes_written, utf16_dst, utf16_dst_length, &status);
+		if (!U_SUCCESS(status)) goto cleanup; /* UTF-16 to UTF-8 conversion failed */
+
+		if (bytes_written >= (int) sizeof(lookupname_buf)) {
+			lookupname = malloc(bytes_written + 1);
+			if (!lookupname) goto cleanup;
+
+			u_strToUTF8(lookupname, bytes_written, NULL, utf16_dst, utf16_dst_length, &status);
+			if (!U_SUCCESS(status)) goto cleanup; /* UTF-16 to UTF-8 conversion failed */
+
+			lookupname[bytes_written] = 0; /* u_strToUTF8() doesn't 0-terminate if dest is filled up */
+		} else {
+			if (!(lookupname = strdup(lookupname)))
+				goto cleanup;
+		}
+
+		if (ascii) {
+			*ascii = lookupname;
+			lookupname = NULL;
+		}
+
+		ret = 0;
+
+cleanup:
+		if (lookupname != lookupname_buf)
+			free(lookupname);
+		if (utf16_src != utf16_src_buf)
+			free(utf16_src);
 	}
 #elif defined(WITH_LIBIDN2)
 #if IDN2_VERSION_NUMBER >= 0x00140000
@@ -772,10 +803,11 @@ static void _add_punycode_if_needed(_psl_idna_t *idna, _psl_vector_t *v, _psl_en
 			_psl_entry_t suffix, *suffixp;
 
 			/* fprintf(stderr, "toASCII '%s' -> '%s'\n", e->label_buf, lookupname); */
-			_suffix_init(&suffix, lookupname, strlen(lookupname));
-			suffix.flags = e->flags;
-			if ((suffixp = _vector_get(v, _vector_add(v, &suffix))))
-				suffixp->label = suffixp->label_buf; /* set label to changed address */
+			if (_suffix_init(&suffix, lookupname, strlen(lookupname)) == 0) {
+				suffix.flags = e->flags;
+				if ((suffixp = _vector_get(v, _vector_add(v, &suffix))))
+					suffixp->label = suffixp->label_buf; /* set label to changed address */
+			}
 		} /* else ignore */
 
 		free(lookupname);
@@ -879,7 +911,7 @@ static int _psl_is_public_suffix(const psl_ctx_t *psl, const char *domain, int t
 		_psl_entry_t *rule = _vector_get(psl->suffixes, 0);
 
 		if (!rule || rule->nlabels < suffix.nlabels - 1)
-			return 0;
+			goto suffix_no;
 
 		rule = _vector_get(psl->suffixes, _vector_find(psl->suffixes, &suffix));
 
@@ -1161,6 +1193,8 @@ psl_ctx_t *psl_load_fp(FILE *fp)
 		/* release unused memory */
 		if ((m = realloc(psl->dafsa, len)))
 			psl->dafsa = m;
+		else if (!len)
+			psl->dafsa = NULL; /* realloc() just free'd psl->dafsa */
 
 		psl->dafsa_size = len;
 		psl->utf8 = !!GetUtfMode(psl->dafsa, len);
@@ -1284,7 +1318,7 @@ void psl_free(psl_ctx_t *psl)
  * psl_builtin:
  *
  * This function returns the PSL context that has been generated and built in at compile-time.
- * You don't have to free the returned context explicitely.
+ * You don't have to free the returned context explicitly.
  *
  * The builtin data also contains punycode entries, one for each international domain name.
  *
@@ -1589,9 +1623,9 @@ int psl_is_cookie_domain_acceptable(const psl_ctx_t *psl, const char *hostname,
  * This helper function converts a string to UTF-8 lowercase + NFKC representation.
  * Lowercase + NFKC UTF-8 is needed as input to the domain checking functions.
  *
- * @lower is set to %NULL on error.
+ * @lower stays unchanged on error.
  *
- * The return value 'lower' must be freed after usage.
+ * When returning PSL_SUCCESS, the return value 'lower' must be freed after usage.
  *
  * Returns: psl_error_t value.
  *   PSL_SUCCESS: Success
@@ -1642,11 +1676,11 @@ psl_error_t psl_str_to_utf8lower(const char *str, const char *encoding _UNUSED,
 		/* C89 allocation */
 		utf16_dst   = alloca(sizeof(UChar) * (str_length * 2 + 1));
 		utf16_lower = alloca(sizeof(UChar) * (str_length * 2 + 1));
-		utf8_lower  = alloca(str_length * 2 + 1);
+		utf8_lower  = alloca(str_length * 6 + 1);
 	} else {
 		utf16_dst   = malloc(sizeof(UChar) * (str_length * 2 + 1));
 		utf16_lower = malloc(sizeof(UChar) * (str_length * 2 + 1));
-		utf8_lower  = malloc(str_length * 2 + 1);
+		utf8_lower  = malloc(str_length * 6 + 1);
 
 		if (!utf16_dst || !utf16_lower || !utf8_lower) {
 			ret = PSL_ERR_NO_MEM;
@@ -1662,21 +1696,16 @@ psl_error_t psl_str_to_utf8lower(const char *str, const char *encoding _UNUSED,
 		if (U_SUCCESS(status)) {
 			int32_t utf16_lower_length = u_strToLower(utf16_lower, str_length * 2 + 1, utf16_dst, utf16_dst_length, locale, &status);
 			if (U_SUCCESS(status)) {
-				u_strToUTF8(utf8_lower, str_length * 8 + 1, NULL, utf16_lower, utf16_lower_length, &status);
+				u_strToUTF8(utf8_lower, str_length * 6 + 1, NULL, utf16_lower, utf16_lower_length, &status);
 				if (U_SUCCESS(status)) {
 					ret = PSL_SUCCESS;
 					if (lower) {
-						if (str_length < 256) {
-							char *tmp = strdup(utf8_lower);
+						char *tmp = strdup(utf8_lower);
 
-							if (tmp)
-								*lower = tmp;
-							else
-								ret = PSL_ERR_NO_MEM;
-						} else {
-							*lower = utf8_lower;
-							utf8_lower = NULL;
-						}
+						if (tmp)
+							*lower = tmp;
+						else
+							ret = PSL_ERR_NO_MEM;
 					}
 				} else {
 					ret = PSL_ERR_TO_UTF8;
@@ -1833,8 +1862,8 @@ psl_ctx_t *psl_latest(const char *fname)
 
 	/* create array of PSL files reverse sorted by mtime (latest first) */
 	ntimes = _insert_file(fname, psl_fname, psl_mtime, 0);
-	ntimes = _insert_file(_psl_filename, psl_fname, psl_mtime, ntimes);
 	ntimes = _insert_file(_psl_dist_filename, psl_fname, psl_mtime, ntimes);
+	ntimes = _insert_file(_psl_filename, psl_fname, psl_mtime, ntimes);
 
 	/* load PSL data from the latest file, falling back to the second recent, ... */
 	for (psl = NULL, it = 0; it < ntimes; it++) {
