@@ -1,5 +1,5 @@
 /*
- * Copyright(c) 2014 Tim Ruehsen
+ * Copyright(c) 2014-2015 Tim Ruehsen
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -50,14 +50,26 @@ static int
 	ok,
 	failed;
 
-static void test(const psl_ctx_t *psl, const char *domain, const char *expected_result)
+static void testx(const psl_ctx_t *psl, const char *domain, const char *encoding, const char *lang, const char *expected_result)
 {
 	const char *result;
-	char *lower;
+	char *lower = NULL;
+	int rc;
 
-	/* our test data is fixed to UTF-8 (english), so provide it here */
-	if (psl_str_to_utf8lower(domain, "utf-8", "en", &lower) == PSL_SUCCESS)
+	/* just to cover special code paths for valgrind checking */
+	psl_str_to_utf8lower(domain, encoding, lang, NULL);
+
+	if ((rc = psl_str_to_utf8lower(domain, encoding, lang, &lower)) == PSL_SUCCESS)
 		domain = lower;
+	/* non-ASCII domains fail here if no runtime IDN library is configured, so skip it */
+#if defined(WITH_LIBIDN) || defined(WITH_LIBIDN2) || defined(WITH_LIBICU)
+	else if (domain) {
+		/* if we do not runtime support, test failure have to be skipped */
+		failed++;
+		printf("psl_str_to_utf8lower(%s)=%d\n", domain ? domain : "NULL", rc);
+		return;
+	}
+#endif
 
 	result = psl_registrable_domain(psl, domain);
 
@@ -66,17 +78,34 @@ static void test(const psl_ctx_t *psl, const char *domain, const char *expected_
 	} else {
 		failed++;
 		printf("psl_registrable_domain(%s)=%s (expected %s)\n",
-			domain, result ? result : "NULL", expected_result ? expected_result : "NULL");
+			   domain ? domain : "NULL", result ? result : "NULL", expected_result ? expected_result : "NULL");
 	}
 
 	free(lower);
 }
 
+static void test(const psl_ctx_t *psl, const char *domain, const char *expected_result)
+{
+	testx(psl, domain, "utf-8", "en", expected_result);
+}
+
+static void test_iso(const psl_ctx_t *psl, const char *domain, const char *expected_result)
+{
+	/* makes only sense with a runtime IDN library configured */
+#if defined(WITH_LIBIDN) || defined(WITH_LIBIDN2) || defined(WITH_LIBICU)
+	testx(psl, domain, "iso-8859-15", "de", expected_result);
+#endif
+}
+
 static void test_psl(void)
 {
 	FILE *fp;
 	const psl_ctx_t *psl;
-	char buf[256], domain[128], expected_regdom[128];
+	const char *p;
+	char buf[256], domain[128], expected_regdom[128], semicolon[2];
+	char lbuf[258];
+	int er_is_null, d_is_null;
+	unsigned it;
 
 	psl = psl_builtin();
 
@@ -99,6 +128,22 @@ static void test_psl(void)
 	/* Norwegian with lowercase oe */
 	test(psl, "www.\303\270yer.no", "www.\303\270yer.no");
 
+	/* Norwegian with lowercase oe, encoded as ISO-8859-15 */
+	test_iso(psl, "www.\370yer.no", "www.\303\270yer.no");
+
+	/* Testing special code paths of psl_str_to_utf8lower() */
+	for (it = 254; it <= 257; it++) {
+		memset(lbuf, 'a', it);
+		lbuf[it] = 0;
+
+		lbuf[0] = '\370';
+		test_iso(psl, lbuf, NULL);
+
+		lbuf[0] = '\303';
+		lbuf[1] = '\270';
+		test(psl, lbuf, NULL);
+	}
+
 	/* special check with NULL psl context and TLD */
 	test(psl, "whoever.forgot.his.name", "whoever.forgot.his.name");
 
@@ -110,15 +155,37 @@ static void test_psl(void)
 
 	if ((fp = fopen(PSL_TESTFILE, "r"))) {
 		while ((fgets(buf, sizeof(buf), fp))) {
-			if (sscanf(buf, " checkPublicSuffix('%127[^']' , '%127[^']", domain, expected_regdom) != 2) {
-				if (sscanf(buf, " checkPublicSuffix('%127[^']' , %127[nul]", domain, expected_regdom) != 2)
+			/* advance over ASCII white space */
+			for (p = buf; *p == ' ' || *p == '\t' || *p == '\r' || *p == '\n'; p++)
+				;
+
+			if (!*p || (*p == '/' && p[1] == '/'))
+				continue; /* ignore comments and blank lines */
+
+			er_is_null = 0;
+			d_is_null = 0;
+
+			if (sscanf(p, "checkPublicSuffix ( '%127[^']' , '%127[^']' ) %1[;]", domain, expected_regdom, semicolon) != 3) {
+				if (sscanf(p, "checkPublicSuffix ( '%127[^']' , null ) %1[;]", domain, semicolon) == 2) {
+					er_is_null = 1;
+				} else if (sscanf(p, "checkPublicSuffix ( null , '%127[^']' ) %1[;]", expected_regdom, semicolon) == 2) {
+					d_is_null = 1;
+				} else if (sscanf(p, "checkPublicSuffix ( null , null ) %1[;]", semicolon) == 1) {
+					d_is_null = 1;
+					er_is_null = 1;
+				} else if (sscanf(p, "%127s %127s", domain, expected_regdom) == 2) {
+					if (!strcmp(domain, "null"))
+						d_is_null = 1;
+					if (!strcmp(expected_regdom, "null"))
+						er_is_null = 1;
+				} else {
+					failed++;
+					printf("Malformed line from '" PSL_TESTFILE "': %s", buf);
 					continue;
+				}
 			}
 
-			if (!strcmp(expected_regdom, "null"))
-				test(psl, domain, NULL);
-			else
-				test(psl, domain, expected_regdom);
+			test(psl, d_is_null ? NULL : domain, er_is_null ? NULL : expected_regdom);
 		}
 
 		fclose(fp);
